#!/usr/bin/env perl
# @(#) A program to download PayPal transactions and save them as QIF files.
#
# Copyright (c) 2015 Dirk Jagdmann <doj@cubic.org>
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
#    1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgement in the product documentation would be
#    appreciated but is not required.
#
#    2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
#
#    3. This notice may not be removed or altered from any source
#    distribution.
#
# For more details visit the website http://llg.cubic.org/csv2iif

use strict;
use Business::PayPal::API qw( GetTransactionDetails TransactionSearch );
use PayPal;
use PayPal::DB;
use Time::Local;
use JSON;
use POSIX;
use QIF;

##############################################################################
# command line
##############################################################################

my $qif;
my $do_help = 0;

my $end_ts = time();
my $start_ts = $end_ts - 86400*10;
if (PayPal::DB::get_state('last_paypal_API_sync') > 0)
{
    $start_ts = PayPal::DB::get_state('last_paypal_API_sync') - 1;
    die "strange start ts $start_ts" unless $start_ts > 100;
}

# parse command line
PayPal::DB::parse_command_line(@ARGV);
foreach my $arg (@ARGV) {
    if ($qif eq 'fetch me') {
	$qif = $arg;
	help() unless length($qif);
	next
    }
    if ($arg eq '-o') {
	$qif = 'fetch me';
	next
    }

    if ($start_ts eq 'fetch me') {
	unless($arg =~ /^(\d{4})-(\d\d)-(\d\d)$/) {
	    die "invalid day specification ($arg) for --start, needs to be 'YYYY-MM-DD'";
	}
	$start_ts = timelocal(0, 0, 0, $3, $2 - 1, $1);
	die "invalid day specification ($arg) for --start" unless $start_ts > 100;
    }
    if ($arg eq '--start') {
	$start_ts = 'fetch me';
	next;
    }

    if ($end_ts eq 'fetch me') {
	unless($arg =~ /^(\d{4})-(\d\d)-(\d\d)$/) {
	    die "invalid day specification ($arg) for --end, needs to be 'YYYY-MM-DD'";
	}
	$end_ts = timelocal(59, 59, 23, $3, $2 - 1, $1);
	die "invalid day specification ($arg) for --end" unless $end_ts > 100;
    }
    if ($arg eq '--end') {
	$end_ts = 'fetch me';
	next;
    }

    if ($arg eq '--lastsync') {
	my $ts = PayPal::DB::get_state('last_paypal_API_sync');
	if ($ts > 0) {
	    print POSIX::ctime($ts);
	} else {
	    print "not synced with PayPal yet.\n";
	}
	exit 0;
    }

    if ($arg eq '--help' || $arg eq '-h') {
	$do_help = 1;
	next;
    }
}

help() if $start_ts eq 'fetch me';
help() if $end_ts eq 'fetch me';
die "start day need to be before end day" if $start_ts > $end_ts;

help() if $do_help;

if (! defined($qif)) {
    my ($s_sec,$s_min,$s_hour,$s_mday,$s_mon,$s_year) = localtime($start_ts);
    my ($e_sec,$e_min,$e_hour,$e_mday,$e_mon,$e_year) = localtime($end_ts);
    $qif = sprintf("PayPal_%04i-%02i-%02iT%02i%02i%02i_%04i-%02i-%02iT%02i%02i%02i.qif",
		   $s_year+1900, $s_mon+1, $s_mday, $s_hour, $s_min, $s_sec,
		   $e_year+1900, $e_mon+1, $e_mday, $e_hour, $e_min, $e_sec,
		  );
}

##############################################################################
# query PayPal
##############################################################################

my $pp = new Business::PayPal::API(
				   Username   => $configuration::paypal_username,
				   Password   => $configuration::paypal_password,
				   Signature  => $configuration::paypal_signature,
				   sandbox    => $configuration::paypal_sandbox
				  ) or die "could not init PayPal API";

my $start_str = PayPal::ts2str($start_ts);
my $end_str = PayPal::ts2str($end_ts);
print "getting PayPal transactions from $start_str to $end_str ...\n";

$| = 1; # auto flush STDOUT
my @txn;
my $total = 0;
while(1)
{
    $end_str = PayPal::ts2str($end_ts);
    #print "# requesting $start_str to $end_str\n\n";
    my $a = $pp->TransactionSearch( StartDate => $start_str, EndDate => $end_str );

    my $z = 0;
    my $last_ts = PayPal::str2ts('2038-01-01T00:00:00Z'); # \todo fix sometime in 2037
    foreach my $h (@$a) {
	++$z;
	++$total;
	my $txn = $h->{TransactionID};
	if (PayPal::DB::get_state($txn)) {
	    print STDERR "skipping $txn from $h->{Timestamp} converted ".asctime(localtime(PayPal::DB::get_state($txn)));
	    next;
	}

	my $ts = PayPal::str2ts($h->{Timestamp});
	die "time stamp is not less than previous time stamp" unless $ts <= $last_ts;
	$last_ts = $ts;

	my %t;
	my $t = PayPal::DB::get_txn($txn);
	if (ref($t) eq 'HASH') {
	    %t = %$t;
	} else {
	    %t = $pp->GetTransactionDetails( TransactionID => $txn );
	}
	# merge $h into %t
	@t{keys %$h} = values %$h;

	#print STDERR "##########################################################\n";
	#print STDERR "# $txn\n";
	#QIF::dump_txn(\%t);

	push(@txn, $txn) if $txn;
	PayPal::DB::save_txn(\%t) or die "$txn could not save PayPal transaction";

	print "\r$total ";
    }

    last if $z < 100;
    die "more than 100 transaction on ".PayPal::ts2str($last_ts) if $end_ts == $last_ts;
    $end_ts = $last_ts + 1; # since there may be multiple transactions on any time stamp, move end_ts out 1 second to catch them all
}
print "\rfetched $total transactions from PayPal.\n";

##############################################################################
# output
##############################################################################

open(A, '>', $qif) or die "could not open for write: $qif";
print A QIF::account('bank', $configuration::qif_paypal_account_name, "PayPal account: $configuration::paypal_login");
print A "!Type:Bank\n";

foreach my $txn (@txn) {
  my $t = PayPal::DB::get_txn($txn);
  die "could not load transaction $txn from disk" unless ref($t) eq 'HASH';

  if ($t->{Type} eq "Payment") {
    if ($t->{NetAmount} < 0.0) {
      QIF::dump_txn($t);
      die "Type Payment expects positive NetAmount";
    }
  } elsif ($t->{Type} eq "Transfer") {
    if ($t->{NetAmount} > 0.0) {
      QIF::dump_txn($t);
      die "Type Transfer expects negative NetAmount";
    }
  } elsif ($t->{Type} eq "Fee Reversal") {
    if ($t->{NetAmount} < 0.0) {
      QIF::dump_txn($t);
      die "Type Fee Reversal expects positive NetAmount";
    }
  } elsif ($t->{Type} eq "Refund") {
    if ($t->{NetAmount} > 0.0) {
      QIF::dump_txn($t);
      die "Type Refund expects negative NetAmount";
    }
  } else {
      QIF::dump_txn($t);
      die "unhandled Type: '$t->{Type}'";
  }

  my $account_name = QIF::lookup_account_name($t);
  unless ($account_name) {
    if ($configuration::handleUnknownItemTitle == configuration::DIE_ON_UNKNOWN_ITEM_TITLE) {
      QIF::dump_txn($t);
      my $k = QIF::txn_type_id($t);
      die "unknown item type: $k";
    } elsif ($configuration::handleUnknownItemTitle == configuration::USE_OTHER_INCOME_ON_UNKNOWN_ITEM_TITLE) {
      $account_name = $configuration::qif_default_income_account_name;
    } elsif ($configuration::handleUnknownItemTitle == configuration::ASK_ON_UNKNOWN_ITEM_TITLE) {
      $account_name = QIF::interactiveAccountSelect($t, $configuration::qif_default_income_account_name, $configuration::qif_paypal_fee_account_name);
      die "could not select account name" unless $account_name;
    } else {
      die "unknown \$configuration::handleUnknownItemTitle value: $configuration::handleUnknownItemTitle";
    }
  }

  $t->{GnuCashAccountName} = $account_name;
  print A QIF::transaction($t, $configuration::qif_paypal_account_name, $configuration::qif_paypal_fee_account_name);
  PayPal::DB::save_txn($t) or die "could not save PayPal transaction";
}

close(A);

PayPal::DB::set_state('last_paypal_API_sync', $end_ts) or die "could not set sync time state";
PayPal::DB::set_states($end_ts, @txn) or die "could not set txn state";
print "saved ".scalar(@txn)." transactions to $qif\n";
exit 0;

##############################################################################

sub help
{
    my $prog = 'paypal2qif.pl';
    print<<"EOT";
usage: $prog
  Fetch transactions from PayPal and save to default filename

usage: $prog -o out.qif
  Fetch transactions from PayPal and convert to out.qif

usage: $prog --start 'YYYY-MM-DD'
  Fetch transactions from PayPal starting at midnight from the day specified.
  Midnight is interpreted in your local time zone. Remember that PayPal uses
  UTC, so the timestamps displayed may have an offset.

usage: $prog --end 'YYYY-MM-DD'
  Fetch transactions from PayPal until 23:59:59 at the day specified.
  Midnight is interpreted in your local time zone. Remember that PayPal uses
  UTC, so the timestamps displayed may have an offset.

usage: $prog --lastsync
  Show timestamp of last sync with PayPal.
EOT
    PayPal::DB::command_line_help($prog);
    exit 1;
}
